/* -*- mode: c; c-file-style: "k&r" -*-

  strnatcmp.c -- Perform 'natural order' comparisons of strings in C.
  Copyright (C) 2000, 2004 by Martin Pool <mbp sourcefrog net>

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
*/


/* partial change history:
 *
 * 2019-07-24 sakir: Added strnnatcmp function to compare non-null-terminated
 *                   nat_char arrays
 *
 * 2012-06-19 barton: Added nat_char_internal so we can perform unsigned comparisons
 *  yet expose a signed API to the external world
 *
 * 2004-10-10 mbp: Lift out character type dependencies into macros.
 *
 * Eric Sosman pointed out that ctype functions take a parameter whose
 * value must be that of an unsigned int, even on platforms that have
 * negative chars in their default char type.
 */

#include <ctype.h>
#include <assert.h>

#include "strnatcmp.h"

typedef	unsigned char nat_char_internal;


/* These are defined as macros to make it easier to adapt this code to
 * different characters types or comparison functions. */
static inline int
nat_isdigit(nat_char_internal a)
{
     return isdigit((unsigned char) a);
}


static inline int
nat_isspace(nat_char_internal a)
{
     return isspace((unsigned char) a);
}


static inline nat_char_internal
nat_toupper(nat_char_internal a)
{
     return toupper((unsigned char) a);
}



static int
compare_right(nat_char_internal const *a, nat_char_internal const *b)
{
     int bias = 0;
     
     /* The longest run of digits wins.  That aside, the greatest
	value wins, but we can't know that it will until we've scanned
	both numbers to know that they have the same magnitude, so we
	remember it in BIAS. */
     for (;; a++, b++) {
	  if (!nat_isdigit(*a)  &&  !nat_isdigit(*b)) {
	       return bias;
	  } else if (!nat_isdigit(*a)) {
	       return -1;
	  } else if (!nat_isdigit(*b)) {
	       return +1;
	  } else if (*a < *b) {
	       if (!bias) {
		    bias = -1;
	       }
	  } else if (*a > *b) {
	       if (!bias) {
		    bias = +1;
	       }
	  } else if (!*a  &&  !*b) {
	       return bias;
	  }
     }

     return 0;
}


static int
compare_left(nat_char_internal const *a, nat_char_internal const *b)
{
     /* Compare two left-aligned numbers: the first to have a
        different value wins. */
     for (;; a++, b++) {
	  if (!nat_isdigit(*a)  &&  !nat_isdigit(*b)) {
	       return 0;
	  } else if (!nat_isdigit(*a)) {
	       return -1;
	  } else if (!nat_isdigit(*b)) {
	       return +1;
	  } else if (*a < *b) {
	       return -1;
	  } else if (*a > *b) {
	       return +1;
	  }
     }
	  
     return 0;
}


static int strnatcmp0(nat_char_internal const *a, nat_char_internal const *b, int fold_case)
{
     int ai, bi;
     nat_char_internal ca, cb;
     int fractional, result;
     
     assert(a && b);
     ai = bi = 0;
     while (1) {
	  ca = a[ai]; cb = b[bi];

	  /* skip over leading spaces or zeros */
	  while (nat_isspace(ca)) {
	       ca = a[++ai];
	  }

	  while (nat_isspace(cb)) {
	       cb = b[++bi];
	  }

	  /* process run of digits */
	  if (nat_isdigit(ca)  &&  nat_isdigit(cb)) {
	       fractional = (ca == '0' || cb == '0');

	       if (fractional) {
		    if ((result = compare_left(a+ai, b+bi)) != 0) {
			 return result;
		    }
	       } else {
		    if ((result = compare_right(a+ai, b+bi)) != 0) {
			 return result;
		    }
	       }
	  }

	  if (!ca && !cb) {
	       /* The strings compare the same.  Perhaps the caller
                  will want to call strcmp to break the tie. */
	       return 0;
	  }

	  if (fold_case) {
	       ca = nat_toupper(ca);
	       cb = nat_toupper(cb);
	  }
	  
	  if (ca < cb) {
	       return -1;
	  } else if (ca > cb) {
	       return +1;
	  }

	  ++ai; ++bi;
     }
}



int strnatcmp(nat_char const *a, nat_char const *b) {
     return strnatcmp0((nat_char_internal const*)a, (nat_char_internal const*)b, 0);
}


/* Compare, recognizing numeric string and ignoring case. */
int strnatcasecmp(nat_char const *a, nat_char const *b) {
     return strnatcmp0((nat_char_internal const*)a, (nat_char_internal const*)b, 1);
}


static int
compare_nright(nat_char_internal const *a, int* ix_a, int len_a,
	       nat_char_internal const *b, int* ix_b, int len_b)
{
     int bias = 0;

     /* The longest run of digits wins.  That aside, the greatest
	value wins, but we can't know that it will until we've scanned
	both numbers to know that they have the same magnitude, so we
	remember it in BIAS. */
     while (1) {
	  nat_char_internal ca = (*ix_a < len_a) ? a[*ix_a] : 0;
	  nat_char_internal cb = (*ix_b < len_b) ? b[*ix_b] : 0;
	  if (!nat_isdigit(ca)) {
           return (nat_isdigit(cb) ? -1 : bias);
	  } else if (!nat_isdigit(cb)) {
	       return +1;
	  } else if (ca < cb) {
	       if (!bias) {
		    bias = -1;
	       }
	  } else if (ca > cb) {
	       if (!bias) {
		    bias = +1;
	       }
	  }
	  // do not increment the indexes beyond the lengths
	  if (*ix_a < len_a) {
	       (*ix_a)++;
	  }
	  if (*ix_b < len_b) {
	       (*ix_b)++;
	  }
     }
     return 0;
}


static int
compare_nleft(nat_char_internal const *a, int* ix_a, int len_a,
	      nat_char_internal const *b, int* ix_b, int len_b)
{
     /* Compare two left-aligned numbers: the first to have a
        different value wins. */
     while (1) {
	  nat_char_internal ca = (*ix_a < len_a) ? a[*ix_a] : 0;
	  nat_char_internal cb = (*ix_b < len_b) ? b[*ix_b] : 0;
	  if (!nat_isdigit(ca)) {
	       return (nat_isdigit(cb)) ? -1 : 0;
	  } else if (!nat_isdigit(cb)) {
	       return +1;
	  } else if (ca < cb) {
	       return -1;
	  } else if (ca > cb) {
	       return 1;
	  }
	  // do not increment the indexes beyond the lengths
	  if (*ix_a < len_a) {
	       (*ix_a)++;
	  }
	  if (*ix_b < len_b) {
	       (*ix_b)++;
	  }
     }
     return 0;
}

static
int strnnatcmp0(nat_char_internal const *a, int len_a, nat_char_internal const *b, int len_b, int fold_case)
{
     int ai, bi;
     int fractional, result;

     if (len_a == 0 && len_b == 0) {
	  return 0;
     }

     assert(a && b);
     ai = bi = 0;
     while (1) {

	  /* skip over leading spaces or zeros */
	  while ((ai < len_a) && nat_isspace(a[ai])) {
	       ai++;
	  }

	  while ((bi < len_b) && nat_isspace(b[bi])) {
	       bi++;
	  }

	  if (ai == len_a && bi == len_b) {
	       return 0;
	  }

	  nat_char_internal ca = (ai < len_a) ? a[ai] : 0;
	  nat_char_internal cb = (bi < len_b) ? b[bi] : 0;

	  /* process run of digits */
	  if (nat_isdigit(ca)  &&  nat_isdigit(cb)) {
	       fractional = (ca == '0' || cb == '0');

	       if (fractional) {
		    if ((result = compare_nleft(a, &ai, len_a,
						b, &bi, len_b)) != 0) {
			 return result;
		    }
	       } else {
		    if ((result = compare_nright(a, &ai, len_a,
						b, &bi, len_b)) != 0) {
			 return result;
		    }
	       }
	       continue;
	  }

	  if (fold_case) {
	       ca = nat_toupper(ca);
	       cb = nat_toupper(cb);
	  }

	  if (ca < cb) {
	       return -1;
	  } else if (ca > cb) {
	       return +1;
	  }

	  // do not increment the indexes beyond the lengths
	  if (ai < len_a) {
	       ++ai;
	  }
	  if (bi < len_b) {
	       ++bi;
	  }
     }
     return 0;
}

int strnnatcmp(nat_char const* a, int len_a,  nat_char const* b, int len_b)
{
     return strnnatcmp0((nat_char_internal const*)a, len_a, (nat_char_internal const*)b, len_b, 0);
}

int strnnatcasecmp(nat_char const* a, int len_a,  nat_char const* b, int len_b)
{
     return strnnatcmp0((nat_char_internal const*)a, len_a, (nat_char_internal const*)b, len_b, 1);
}
